@@ -26,6 +26,20 @@
 
 -define(SERVER, ?MODULE).
 
+%%----------------------------------------------------------------------------
+
+-ifdef(use_specs).
+
+-spec(start_link/0 :: () -> rabbit_types:ok_pid_or_error()).
+-spec(start_child/2 ::
+        (node(), [any()]) -> rabbit_types:ok(pid() | undefined) |
+                             rabbit_types:ok({pid(), any()}) |
+                             rabbit_types:error(any())).
+
+-endif.
+
+%%----------------------------------------------------------------------------
+
 start_link() ->
     supervisor2:start_link({local, ?SERVER}, ?MODULE, []).
 
