@@ -24,7 +24,7 @@
 
 -behaviour(gen_server2).
 
--export([start_link/0, update/0, register/2, deregister/1,
+-export([start_link/0, register/2, deregister/1,
          report_ram_duration/2, stop/0]).
 
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
@@ -36,7 +36,6 @@
                 queue_durations,      %% ets #process
                 queue_duration_sum,   %% sum of all queue_durations
                 queue_duration_count, %% number of elements in sum
-                memory_limit,         %% how much memory we intend to use
                 desired_duration      %% the desired queue duration
                }).
 
@@ -63,9 +62,6 @@
 -define(SUM_INC_THRESHOLD, 0.95).
 -define(SUM_INC_AMOUNT, 1.0).
 
-%% If user disabled vm_memory_monitor, let's assume 1GB of memory we can use.
--define(MEMORY_SIZE_FOR_DISABLED_VMM, 1073741824).
-
 -define(EPSILON, 0.000001). %% less than this and we clamp to 0
 
 %%----------------------------------------------------------------------------
@@ -73,7 +69,6 @@
 -ifdef(use_specs).
 
 -spec(start_link/0 :: () -> rabbit_types:ok_pid_or_error()).
--spec(update/0 :: () -> 'ok').
 -spec(register/2 :: (pid(), {atom(),atom(),[any()]}) -> 'ok').
 -spec(deregister/1 :: (pid()) -> 'ok').
 -spec(report_ram_duration/2 ::
@@ -89,9 +84,6 @@
 start_link() ->
     gen_server2:start_link({local, ?SERVER}, ?MODULE, [], []).
 
-update() ->
-    gen_server2:cast(?SERVER, update).
-
 register(Pid, MFA = {_M, _F, _A}) ->
     gen_server2:call(?SERVER, {register, Pid, MFA}, infinity).
 
@@ -110,15 +102,7 @@ stop() ->
 %%----------------------------------------------------------------------------
 
 init([]) ->
-    MemoryLimit = trunc(?MEMORY_LIMIT_SCALING *
-                            (try
-                                 vm_memory_monitor:get_memory_limit()
-                             catch
-                                 exit:{noproc, _} -> ?MEMORY_SIZE_FOR_DISABLED_VMM
-                             end)),
-
-    {ok, TRef} = timer:apply_interval(?DEFAULT_UPDATE_INTERVAL,
-                                      ?SERVER, update, []),
+    {ok, TRef} = timer:send_interval(?DEFAULT_UPDATE_INTERVAL, update),
 
     Ets = ets:new(?TABLE_NAME, [set, private, {keypos, #process.pid}]),
 
@@ -127,7 +111,6 @@ init([]) ->
                     queue_durations      = Ets,
                     queue_duration_sum   = 0.0,
                     queue_duration_count = 0,
-                    memory_limit         = MemoryLimit,
                     desired_duration     = infinity })}.
 
 handle_call({report_ram_duration, Pid, QueueDuration}, From,
@@ -165,9 +148,6 @@ handle_call({register, Pid, MFA}, _From,
 handle_call(_Request, _From, State) ->
     {noreply, State}.
 
-handle_cast(update, State) ->
-    {noreply, internal_update(State)};
-
 handle_cast({deregister, Pid}, State) ->
     {noreply, internal_deregister(Pid, true, State)};
 
@@ -177,6 +157,9 @@ handle_cast(stop, State) ->
 handle_cast(_Request, State) ->
     {noreply, State}.
 
+handle_info(update, State) ->
+    {noreply, internal_update(State)};
+
 handle_info({'DOWN', _MRef, process, Pid, _Reason}, State) ->
     {noreply, internal_deregister(Pid, false, State)};
 
@@ -223,22 +206,24 @@ internal_deregister(Pid, Demonitor,
                            queue_duration_count = Count1 }
     end.
 
-internal_update(State = #state { memory_limit = Limit,
-                                 queue_durations = Durations,
+internal_update(State = #state { queue_durations = Durations,
                                  desired_duration = DesiredDurationAvg,
                                  queue_duration_sum = Sum,
                                  queue_duration_count = Count }) ->
-    MemoryRatio = erlang:memory(total) / Limit,
+    MemoryLimit = ?MEMORY_LIMIT_SCALING * vm_memory_monitor:get_memory_limit(),
+    MemoryRatio = case MemoryLimit > 0.0 of
+                      true  -> erlang:memory(total) / MemoryLimit;
+                      false -> infinity
+                  end,
     DesiredDurationAvg1 =
-        case MemoryRatio < ?LIMIT_THRESHOLD orelse Count == 0 of
-            true ->
+        if MemoryRatio =:= infinity ->
+                0.0;
+           MemoryRatio < ?LIMIT_THRESHOLD orelse Count == 0 ->
                 infinity;
-            false ->
-                Sum1 = case MemoryRatio < ?SUM_INC_THRESHOLD of
-                           true  -> Sum + ?SUM_INC_AMOUNT;
-                           false -> Sum
-                       end,
-                (Sum1 / Count) / MemoryRatio
+           MemoryRatio < ?SUM_INC_THRESHOLD ->
+                ((Sum + ?SUM_INC_AMOUNT) / Count) / MemoryRatio;
+           true ->
+                (Sum / Count) / MemoryRatio
         end,
     State1 = State #state { desired_duration = DesiredDurationAvg1 },
 
