@@ -45,8 +45,19 @@
 -behaviour(gm).
 
 -include("rabbit.hrl").
+
+%%----------------------------------------------------------------------------
+
 -include("gm_specs.hrl").
 
+-ifdef(use_specs).
+%% Shut dialyzer up
+-spec(promote_me/2 :: (_, _) -> no_return()).
+-endif.
+
+%%----------------------------------------------------------------------------
+
+
 -define(CREATION_EVENT_KEYS,
         [pid,
          name,
@@ -379,9 +390,9 @@ needs_confirming(_Delivery, _State) ->
     immediately.
 
 confirm_messages(MsgIds, State = #state { msg_id_status = MS }) ->
-    {MS1, CMs} =
+    {CMs, MS1} =
         lists:foldl(
-          fun (MsgId, {MSN, CMsN} = Acc) ->
+          fun (MsgId, {CMsN, MSN} = Acc) ->
                   %% We will never see 'discarded' here
                   case dict:find(MsgId, MSN) of
                       error ->
@@ -391,12 +402,12 @@ confirm_messages(MsgIds, State = #state { msg_id_status = MS }) ->
                       {ok, {published, ChPid}} ->
                           %% Still not seen it from the channel, just
                           %% record that it's been confirmed.
-                          {dict:store(MsgId, {confirmed, ChPid}, MSN), CMsN};
+                          {CMsN, dict:store(MsgId, {confirmed, ChPid}, MSN)};
                       {ok, {published, ChPid, MsgSeqNo}} ->
                           %% Seen from both GM and Channel. Can now
                           %% confirm.
-                          {dict:erase(MsgId, MSN),
-                           gb_trees_cons(ChPid, MsgSeqNo, CMsN)};
+                          {rabbit_misc:gb_trees_cons(ChPid, MsgSeqNo, CMsN),
+                           dict:erase(MsgId, MSN)};
                       {ok, {confirmed, _ChPid}} ->
                           %% It's already been confirmed. This is
                           %% probably it's been both sync'd to disk
@@ -405,17 +416,10 @@ confirm_messages(MsgIds, State = #state { msg_id_status = MS }) ->
                           %% channel. Nothing to do here.
                           Acc
                   end
-          end, {MS, gb_trees:empty()}, MsgIds),
-    [ok = rabbit_channel:confirm(ChPid, MsgSeqNos)
-     || {ChPid, MsgSeqNos} <- gb_trees:to_list(CMs)],
+          end, {gb_trees:empty(), MS}, MsgIds),
+    rabbit_misc:gb_trees_foreach(fun rabbit_channel:confirm/2, CMs),
     State #state { msg_id_status = MS1 }.
 
-gb_trees_cons(Key, Value, Tree) ->
-    case gb_trees:lookup(Key, Tree) of
-        {value, Values} -> gb_trees:update(Key, [Value | Values], Tree);
-        none            -> gb_trees:insert(Key, [Value], Tree)
-    end.
-
 handle_process_result({ok,   State}) -> noreply(State);
 handle_process_result({stop, State}) -> {stop, normal, State}.
 
@@ -522,9 +526,11 @@ promote_me(From, #state { q                   = Q = #amqqueue { name = QName },
     MasterState = rabbit_mirror_queue_master:promote_backing_queue_state(
                     CPid, BQ, BQS, GM, SS, MonitoringPids),
 
-    MTC = dict:from_list(
-            [{MsgId, {ChPid, MsgSeqNo}} ||
-                {MsgId, {published, ChPid, MsgSeqNo}} <- dict:to_list(MS)]),
+    MTC = lists:foldl(fun ({MsgId, {published, ChPid, MsgSeqNo}}, MTC0) ->
+                              gb_trees:insert(MsgId, {ChPid, MsgSeqNo}, MTC0);
+                          (_, MTC0) ->
+                              MTC0
+                      end, gb_trees:empty(), MSList),
     NumAckTags = [NumAckTag || {_MsgId, NumAckTag} <- dict:to_list(MA)],
     AckTags = [AckTag || {_Num, AckTag} <- lists:sort(NumAckTags)],
     Deliveries = [Delivery || {_ChPid, {PubQ, _PendCh}} <- dict:to_list(SQ),
@@ -721,7 +727,7 @@ process_instruction(
                     never ->
                         {MQ2, PendingCh, MS};
                     eventually ->
-                        {MQ2, sets:add_element(MsgId, PendingCh),
+                        {MQ2, PendingCh,
                          dict:store(MsgId, {published, ChPid, MsgSeqNo}, MS)};
                     immediately ->
                         ok = rabbit_channel:confirm(ChPid, [MsgSeqNo]),
@@ -823,14 +829,14 @@ process_instruction({ack, MsgIds},
     [] = MsgIds1 -- MsgIds, %% ASSERTION
     {ok, State #state { msg_id_ack          = MA1,
                         backing_queue_state = BQS1 }};
-process_instruction({requeue, MsgPropsFun, MsgIds},
+process_instruction({requeue, MsgIds},
                     State = #state { backing_queue       = BQ,
                                      backing_queue_state = BQS,
                                      msg_id_ack          = MA }) ->
     {AckTags, MA1} = msg_ids_to_acktags(MsgIds, MA),
     {ok, case length(AckTags) =:= length(MsgIds) of
              true ->
-                 {MsgIds, BQS1} = BQ:requeue(AckTags, MsgPropsFun, BQS),
+                 {MsgIds, BQS1} = BQ:requeue(AckTags, BQS),
                  State #state { msg_id_ack          = MA1,
                                 backing_queue_state = BQS1 };
              false ->
