@@ -78,6 +78,33 @@
 -spec(on_node_down/1 :: (node()) -> 'ok').
 -spec(check_tcp_listener_address/2 :: (atom(), listener_config())
         -> [{inet:ip_address(), ip_port(), family(), atom()}]).
+-spec(ensure_ssl/0 :: () -> rabbit_types:infos()).
+-spec(ssl_transform_fun/1 ::
+        (rabbit_types:infos())
+        -> fun ((rabbit_net:socket())
+                -> rabbit_types:ok_or_error(#ssl_socket{}))).
+
+-spec(boot/0 :: () -> 'ok').
+-spec(start_client/1 ::
+	(port() | #ssl_socket{ssl::{'sslsocket',_,_}}) ->
+			     atom() | pid() | port() | {atom(),atom()}).
+-spec(start_ssl_client/2 ::
+	(_,port() | #ssl_socket{ssl::{'sslsocket',_,_}}) ->
+				 atom() | pid() | port() | {atom(),atom()}).
+-spec(tcp_listener_started/3 ::
+	(_,
+         string() |
+	 {byte(),byte(),byte(),byte()} |
+	 {char(),char(),char(),char(),char(),char(),char(),char()},
+	 _) ->
+				     'ok').
+-spec(tcp_listener_stopped/3 ::
+	(_,
+         string() |
+	 {byte(),byte(),byte(),byte()} |
+	 {char(),char(),char(),char(),char(),char(),char(),char()},
+	 _) ->
+				     'ok').
 
 -endif.
 
@@ -280,9 +307,15 @@ connections() ->
                                      rabbit_networking, connections_local, []).
 
 connections_local() ->
-    [rabbit_connection_sup:reader(ConnSup) ||
+    [Reader ||
         {_, ConnSup, supervisor, _}
-            <- supervisor:which_children(rabbit_tcp_client_sup)].
+            <- supervisor:which_children(rabbit_tcp_client_sup),
+        Reader <- [try
+                       rabbit_connection_sup:reader(ConnSup)
+                   catch exit:{noproc, _} ->
+                           noproc
+                   end],
+        Reader =/= noproc].
 
 connection_info_keys() -> rabbit_reader:info_keys().
 
@@ -293,6 +326,7 @@ connection_info_all() -> cmap(fun (Q) -> connection_info(Q) end).
 connection_info_all(Items) -> cmap(fun (Q) -> connection_info(Q, Items) end).
 
 close_connection(Pid, Explanation) ->
+    rabbit_log:info("Closing connection ~p because ~p~n", [Pid, Explanation]),
     case lists:member(Pid, connections()) of
         true  -> rabbit_reader:shutdown(Pid, Explanation);
         false -> throw({error, {not_a_connection_pid, Pid}})
