@@ -32,7 +32,7 @@
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          terminate/2, code_change/3]).
 
--export([update/0, get_total_memory/0, get_vm_limit/0,
+-export([get_total_memory/0, get_vm_limit/0,
          get_check_interval/0, set_check_interval/1,
          get_vm_memory_high_watermark/0, set_vm_memory_high_watermark/1,
          get_memory_limit/0]).
@@ -40,6 +40,7 @@
 
 -define(SERVER, ?MODULE).
 -define(DEFAULT_MEMORY_CHECK_INTERVAL, 1000).
+-define(ONE_MB, 1048576).
 
 %% For an unknown OS, we assume that we have 1GB of memory. It'll be
 %% wrong. Scale by vm_memory_high_watermark in configuration to get a
@@ -57,15 +58,14 @@
 
 -ifdef(use_specs).
 
--spec(start_link/1 :: (float()) -> {'ok', pid()} | {'error', any()}).
--spec(update/0 :: () -> 'ok').
+-spec(start_link/1 :: (float()) -> rabbit_types:ok_pid_or_error()).
 -spec(get_total_memory/0 :: () -> (non_neg_integer() | 'unknown')).
 -spec(get_vm_limit/0 :: () -> non_neg_integer()).
--spec(get_memory_limit/0 :: () -> non_neg_integer()).
 -spec(get_check_interval/0 :: () -> non_neg_integer()).
 -spec(set_check_interval/1 :: (non_neg_integer()) -> 'ok').
 -spec(get_vm_memory_high_watermark/0 :: () -> float()).
 -spec(set_vm_memory_high_watermark/1 :: (float()) -> 'ok').
+-spec(get_memory_limit/0 :: () -> non_neg_integer()).
 
 -endif.
 
@@ -73,9 +73,6 @@
 %% Public API
 %%----------------------------------------------------------------------------
 
-update() ->
-    gen_server:cast(?SERVER, update).
-
 get_total_memory() ->
     get_total_memory(os:type()).
 
@@ -106,35 +103,20 @@ start_link(Args) ->
     gen_server:start_link({local, ?SERVER}, ?MODULE, [Args], []).
 
 init([MemFraction]) ->
-    TotalMemory =
-        case get_total_memory() of
-            unknown ->
-                error_logger:warning_msg(
-                  "Unknown total memory size for your OS ~p. "
-                  "Assuming memory size is ~pMB.~n",
-                  [os:type(), trunc(?MEMORY_SIZE_FOR_UNKNOWN_OS/1048576)]),
-                ?MEMORY_SIZE_FOR_UNKNOWN_OS;
-            M -> M
-        end,
-    MemLimit = get_mem_limit(MemFraction, TotalMemory),
-    error_logger:info_msg("Memory limit set to ~pMB.~n",
-                          [trunc(MemLimit/1048576)]),
     TRef = start_timer(?DEFAULT_MEMORY_CHECK_INTERVAL),
-    State = #state { total_memory = TotalMemory,
-                     memory_limit = MemLimit,
-                     timeout = ?DEFAULT_MEMORY_CHECK_INTERVAL,
+    State = #state { timeout = ?DEFAULT_MEMORY_CHECK_INTERVAL,
                      timer = TRef,
                      alarmed = false},
-    {ok, internal_update(State)}.
+    {ok, set_mem_limits(State, MemFraction)}.
 
 handle_call(get_vm_memory_high_watermark, _From, State) ->
     {reply, State#state.memory_limit / State#state.total_memory, State};
 
 handle_call({set_vm_memory_high_watermark, MemFraction}, _From, State) ->
-    MemLimit = get_mem_limit(MemFraction, State#state.total_memory),
+    State1 = set_mem_limits(State, MemFraction),
     error_logger:info_msg("Memory alarm changed to ~p, ~p bytes.~n",
-                          [MemFraction, MemLimit]),
-    {reply, ok, State#state{memory_limit = MemLimit}};
+                          [MemFraction, State1#state.memory_limit]),
+    {reply, ok, State1};
 
 handle_call(get_check_interval, _From, State) ->
     {reply, State#state.timeout, State};
@@ -149,12 +131,12 @@ handle_call(get_memory_limit, _From, State) ->
 handle_call(_Request, _From, State) ->
     {noreply, State}.
 
-handle_cast(update, State) ->
-    {noreply, internal_update(State)};
-
 handle_cast(_Request, State) ->
     {noreply, State}.
 
+handle_info(update, State) ->
+    {noreply, internal_update(State)};
+
 handle_info(_Info, State) ->
     {noreply, State}.
 
@@ -168,6 +150,30 @@ code_change(_OldVsn, State, _Extra) ->
 %% Server Internals
 %%----------------------------------------------------------------------------
 
+set_mem_limits(State, MemFraction) ->
+    TotalMemory =
+        case get_total_memory() of
+            unknown ->
+                case State of
+                    #state { total_memory = undefined,
+                             memory_limit = undefined } ->
+                        error_logger:warning_msg(
+                          "Unknown total memory size for your OS ~p. "
+                          "Assuming memory size is ~pMB.~n",
+                          [os:type(),
+                           trunc(?MEMORY_SIZE_FOR_UNKNOWN_OS/?ONE_MB)]);
+                    _ ->
+                        ok
+                end,
+                ?MEMORY_SIZE_FOR_UNKNOWN_OS;
+            M -> M
+        end,
+    MemLim = get_mem_limit(MemFraction, TotalMemory),
+    error_logger:info_msg("Memory limit set to ~pMB of ~pMB total.~n",
+                          [trunc(MemLim/?ONE_MB), trunc(TotalMemory/?ONE_MB)]),
+    internal_update(State #state { total_memory = TotalMemory,
+                                   memory_limit = MemLim }).
+
 internal_update(State = #state { memory_limit = MemLimit,
                                  alarmed = Alarmed}) ->
     MemUsed = erlang:memory(total),
@@ -190,7 +196,7 @@ emit_update_info(State, MemUsed, MemLimit) ->
       [State, MemUsed, MemLimit]).
 
 start_timer(Timeout) ->
-    {ok, TRef} = timer:apply_interval(Timeout, ?MODULE, update, []),
+    {ok, TRef} = timer:send_interval(Timeout, update),
     TRef.
 
 %% According to http://msdn.microsoft.com/en-us/library/aa366778(VS.85).aspx
@@ -322,9 +328,9 @@ parse_line_sunos(Line) ->
             [Value1 | UnitsRest] = string:tokens(RHS, " "),
             Value2 = case UnitsRest of
                          ["Gigabytes"] ->
-                             list_to_integer(Value1) * 1024 * 1024 * 1024;
+                             list_to_integer(Value1) * ?ONE_MB * 1024;
                          ["Megabytes"] ->
-                             list_to_integer(Value1) * 1024 * 1024;
+                             list_to_integer(Value1) * ?ONE_MB;
                          ["Kilobytes"] ->
                              list_to_integer(Value1) * 1024;
                          _ ->
